/*
 * Copyright (C) 2002 Nikos Mavroyanopoulos
 *
 * This file is part of Hydra webserver.
 *
 * Hydra is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * Hydra is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "boa.h"

#ifdef ENABLE_SSL

#include <gnutls/gnutls.h>
#include <gcrypt.h>

#ifdef ENABLE_SMP
pthread_mutex_t ssl_session_cache_lock = PTHREAD_MUTEX_INITIALIZER;
#endif

extern int ssl_session_cache;
extern int ssl_session_timeout;

extern char* ssl_ciphers;
extern char* ssl_kx;
extern char* ssl_mac;
extern char* ssl_comp;
extern char* ssl_protocol;

static void wrap_db_init(void);
static int wrap_db_store(void *dbf, gnutls_datum key, gnutls_datum data);
static gnutls_datum wrap_db_fetch(void *dbf, gnutls_datum key);
static int wrap_db_delete(void *dbf, gnutls_datum key);

static int cur = 0; /* points to the credentials structure used */
static gnutls_certificate_credentials credentials[2];

static int need_dh_params = 0; /* whether we need to generate DHE
 * parameters. Depend on the chosen ciphersuites.
 */
static int need_rsa_params = 0;


/* we use primes up to 1024 in this server.
 * otherwise we should add them here.
 */
extern int ssl_dh_bits;

gnutls_dh_params _dh_params[2];
gnutls_rsa_params _rsa_params[2];

static int generate_dh_primes( gnutls_dh_params* dh_params)
{
    gnutls_datum prime, generator;
    
    if (gnutls_dh_params_init( dh_params) < 0) {
        log_error_time();
	fprintf(stderr, "Error in dh parameter initialization\n");
	exit(1);
    }

    /* Generate Diffie Hellman parameters - for use with DHE
     * kx algorithms. These should be discarded and regenerated
     * once a day, once a week or once a month. Depends on the
     * security requirements.
     */

     if (gnutls_dh_params_generate(&prime, &generator, ssl_dh_bits) <
	    0) {
	    log_error_time();
	    fprintf(stderr, "Error in prime generation\n");
	    exit(1);
     }

     if (gnutls_dh_params_set
	    (*dh_params, prime, generator, ssl_dh_bits) < 0) {
	    log_error_time();
	    fprintf(stderr, "Error in prime replacement\n");
	    exit(1);
     }

     log_error_time();
     fprintf
	    (stderr,
	     "tls: Generated Diffie Hellman parameters [%d bits].\n",
	     ssl_dh_bits);

     free(prime.data);
     free(generator.data);

    return 0;
}

static int generate_rsa_params( gnutls_rsa_params* rsa_params)
{
    gnutls_datum m, e, d, p, q, u;

    if (gnutls_rsa_params_init( rsa_params) < 0) {
	log_error_time();
	fprintf(stderr, "Error in rsa parameter initialization\n");
	exit(1);
    }

    /* Generate RSA parameters - for use with RSA-export
     * cipher suites. These should be discarded and regenerated
     * once a day, once every 500 transactions etc. Depends on the
     * security requirements.
     */

    if (gnutls_rsa_params_generate(&m, &e, &d, &p, &q, &u, 512) < 0) {
	log_error_time();
	fprintf(stderr, "Error in rsa parameter generation\n");
	exit(1);
    }

    if (gnutls_rsa_params_set( *rsa_params, m, e, d, p, q, u, 512) < 0) {
	log_error_time();
	fprintf(stderr, "Error in rsa parameter setting\n");
	exit(1);
    }

    free(m.data);
    free(e.data);
    free(d.data);
    free(p.data);
    free(q.data);
    free(u.data);

    log_error_time();
    fprintf
	(stderr, "tls: Generated temporary RSA parameters.\n");

    return 0;
}

static int protocol_priority[16];
static int kx_priority[16];
static int cipher_priority[16];
static int mac_priority[16];
static int comp_priority[16];

/* Parses a string in the form:
 * CIPHER1, CIPHER2, ...
 * and tries to find the given algorithm.
 * Returns true or false.
 */
static int parse_cs_string( const char* string, const char* algo)
{
	if (string == NULL || algo == NULL) return 0;

	if (strstr( string, algo) != NULL)
		return 1;
		
	return 0;

}

gnutls_session initialize_ssl_session(void)
{
    GNUTLS_STATE state;
    
    gnutls_init(&state, GNUTLS_SERVER);

    gnutls_cipher_set_priority(state, cipher_priority);
    gnutls_compression_set_priority(state, comp_priority);
    gnutls_kx_set_priority(state, kx_priority);
    gnutls_protocol_set_priority(state, protocol_priority);
    gnutls_mac_set_priority(state, mac_priority);

    gnutls_cred_set(state, GNUTLS_CRD_CERTIFICATE, credentials[ cur]);

    gnutls_certificate_server_set_request(state, GNUTLS_CERT_IGNORE);

    if (ssl_session_cache != 0) {
	gnutls_db_set_retrieve_function(state, wrap_db_fetch);
	gnutls_db_set_remove_function(state, wrap_db_delete);
	gnutls_db_set_store_function(state, wrap_db_store);
	gnutls_db_set_ptr(state, NULL);
    }
    gnutls_db_set_cache_expiration( state, ssl_session_timeout);

    gnutls_handshake_set_private_extensions( state, 1);

    return state;
}

extern char *server_cert;
extern char *server_key;

/* Initialization of gnutls' global state
 */
int initialize_ssl(void)
{
    int i;

    gnutls_global_init();
/*    gcry_control (GCRYCTL_DISABLE_INTERNAL_LOCKING, NULL, 0); */

    if (gnutls_certificate_allocate_credentials( &credentials[0]) < 0) {
	log_error_time();
	fprintf(stderr, "certificate allocation error\n");
	exit(1);
    }

    if (gnutls_certificate_set_x509_key_file
	( credentials[0], server_cert, server_key, GNUTLS_X509_FMT_PEM) < 0) {
	log_error_time();
	fprintf(stderr, "could not find %s or %s", server_cert,
		server_key);
	exit(1);
    }

    if (ssl_session_cache != 0)
	wrap_db_init();

    /* Add ciphers 
     */
    i = 0;
    if ( parse_cs_string( ssl_ciphers, "AES") != 0)
    	cipher_priority[i++] = GNUTLS_CIPHER_RIJNDAEL_128_CBC;
    if ( parse_cs_string( ssl_ciphers, "ARCFOUR-128") != 0)
    	cipher_priority[i++] = GNUTLS_CIPHER_ARCFOUR_128;
    if ( parse_cs_string( ssl_ciphers, "3DES") != 0)
    	cipher_priority[i++] = GNUTLS_CIPHER_3DES_CBC;
    if ( parse_cs_string( ssl_ciphers, "ARCFOUR-40") != 0)
    	cipher_priority[i++] = GNUTLS_CIPHER_ARCFOUR_40;
    cipher_priority[i] = 0;

    /* Add key exchange methods
     */
    i = 0;
    if ( parse_cs_string( ssl_kx, "RSA") != 0)
    	kx_priority[i++] = GNUTLS_KX_RSA;
    if ( parse_cs_string( ssl_kx, "RSA-EXPORT") != 0) {
    	kx_priority[i++] = GNUTLS_KX_RSA_EXPORT;
    	need_rsa_params = 1;
    }
    if ( parse_cs_string( ssl_kx, "DHE-RSA") != 0) {
    	kx_priority[i++] = GNUTLS_KX_DHE_RSA;
    	need_dh_params = 1; /* generate DH parameters */
    }
    if ( parse_cs_string( ssl_kx, "DHE-DSS") != 0) {
    	kx_priority[i++] = GNUTLS_KX_DHE_DSS;
    	need_dh_params = 1;
    }
    kx_priority[i] = 0;

    /* Add MAC Algorithms
     */
    i = 0;
    if ( parse_cs_string( ssl_mac, "MD5") != 0)
    	mac_priority[i++] = GNUTLS_MAC_MD5;
    if ( parse_cs_string( ssl_mac, "SHA") != 0)
    	mac_priority[i++] = GNUTLS_MAC_SHA;
    mac_priority[i] = 0;

    /* Add Compression algorithms
     */
    i = 0;
    if ( parse_cs_string( ssl_comp, "NULL") != 0)
    	comp_priority[i++] = GNUTLS_COMP_NULL;
    if ( parse_cs_string( ssl_comp, "ZLIB") != 0)
    	comp_priority[i++] = GNUTLS_COMP_ZLIB;
    if ( parse_cs_string( ssl_comp, "LZO") != 0)
    	comp_priority[i++] = GNUTLS_COMP_LZO;
    comp_priority[i] = 0;

    /* Add protocols
     */
    i = 0;
    if ( parse_cs_string( ssl_protocol, "TLS") != 0)
    	protocol_priority[i++] = GNUTLS_TLS1;
    if ( parse_cs_string( ssl_protocol, "SSL") != 0)
    	protocol_priority[i++] = GNUTLS_SSL3;
    protocol_priority[i] = 0;

    /* Generate temporary parameters -- if needed.
     */
    if (need_rsa_params) {
    	generate_rsa_params( &_rsa_params[0]);
	gnutls_certificate_set_rsa_params(credentials[0], _rsa_params[0]);
    }

    if (need_dh_params) {
	generate_dh_primes( &_dh_params[0]);
	gnutls_certificate_set_dh_params(credentials[0], _dh_params[0]);
    }

    return 0;
}

/* This function will regenerate the SSL parameters (RSA and DH) without
 * any need for downtime.
 */
void ssl_regenerate_params(void)
{
int _cur = (cur + 1) % 2;

/* The hint here, is that we keep a copy of 2 certificate credentials.
 * When we come here, we free the unused copy and allocate new
 * parameters to it. Then we make the current copy to be this copy.
 *
 * We don't free the previous copy because we don't know if anyone
 * is using it. (this has to be fixed)
 */

    time(&current_time);

    if ( !credentials[_cur]) {
       if (gnutls_certificate_allocate_credentials( &credentials[ _cur]) < 0) {
 	  log_error_time();
 	  fprintf(stderr, "certificate allocation error\n");
	  exit(1);
       }

       if (gnutls_certificate_set_x509_key_file
  	   ( credentials[_cur], server_cert, server_key, GNUTLS_X509_FMT_PEM) < 0) {
	   log_error_time();
	   fprintf(stderr, "could not find %s or %s", server_cert,
		server_key);
	   exit(1);
       }
    }
    
    if (need_rsa_params) {
	gnutls_rsa_params_deinit( _rsa_params[ _cur]);
	generate_rsa_params( &_rsa_params[ _cur]);
        gnutls_certificate_set_rsa_params(credentials[_cur], _rsa_params[ _cur]);
    }

    if (need_dh_params) {
	gnutls_dh_params_deinit( _dh_params[ _cur]);
        generate_dh_primes( &_dh_params[ _cur]);
        gnutls_certificate_set_dh_params(credentials[_cur], _dh_params[ _cur]);
    }

    cur = _cur;

    return;
}


/* Session resuming: 
 */

#define SESSION_ID_SIZE 32
#define SESSION_DATA_SIZE 512

typedef struct {
    char session_id[SESSION_ID_SIZE];
    int session_id_size;

    char session_data[SESSION_DATA_SIZE];
    int session_data_size;
} CACHE;

static CACHE *cache_db;
static int cache_db_ptr;

static void wrap_db_init(void)
{

    /* allocate cache_db */
    cache_db = calloc(1, ssl_session_cache * sizeof(CACHE));
}

static int wrap_db_store(void *dbf, gnutls_datum key, gnutls_datum data)
{

    if (cache_db == NULL)
	return -1;

    if (key.size > SESSION_ID_SIZE)
	return -1;
    if (data.size > SESSION_DATA_SIZE)
	return -1;

#ifdef ENABLE_SMP
    pthread_mutex_lock( &ssl_session_cache_lock);
#endif

    memcpy(cache_db[cache_db_ptr].session_id, key.data, key.size);
    cache_db[cache_db_ptr].session_id_size = key.size;

    memcpy(cache_db[cache_db_ptr].session_data, data.data, data.size);
    cache_db[cache_db_ptr].session_data_size = data.size;

    cache_db_ptr++;
    cache_db_ptr %= ssl_session_cache;

#ifdef ENABLE_SMP
    pthread_mutex_unlock( &ssl_session_cache_lock);
#endif

    return 0;
}

static gnutls_datum wrap_db_fetch(void *dbf, gnutls_datum key)
{
    gnutls_datum res = { NULL, 0 };
    int i;

    if (cache_db == NULL)
	return res;

#ifdef ENABLE_SMP
    pthread_mutex_lock( &ssl_session_cache_lock);
#endif

    for (i = 0; i < ssl_session_cache; i++) {
	if (key.size == cache_db[i].session_id_size &&
	    memcmp(key.data, cache_db[i].session_id, key.size) == 0) {

	    res.size = cache_db[i].session_data_size;

	    res.data = malloc(res.size);
	    if (res.data == NULL) {
#ifdef ENABLE_SMP
                pthread_mutex_unlock( &ssl_session_cache_lock);
#endif
		return res;
            }

	    memcpy(res.data, cache_db[i].session_data, res.size);

#ifdef ENABLE_SMP
            pthread_mutex_unlock( &ssl_session_cache_lock);
#endif
	    return res;
	}
    }

#ifdef ENABLE_SMP
    pthread_mutex_unlock( &ssl_session_cache_lock);
#endif

    return res;
}

static int wrap_db_delete(void *dbf, gnutls_datum key)
{
int i;

    if (cache_db == NULL)
	return -1;

#ifdef ENABLE_SMP
    pthread_mutex_lock( &ssl_session_cache_lock);
#endif

    for (i = 0; i < ssl_session_cache; i++) {
	if (key.size == cache_db[i].session_id_size &&
	    memcmp(key.data, cache_db[i].session_id, key.size) == 0) {

	    cache_db[i].session_id_size = 0;
	    cache_db[i].session_data_size = 0;

#ifdef ENABLE_SMP
            pthread_mutex_unlock( &ssl_session_cache_lock);
#endif

	    return 0;
	}
    }

#ifdef ENABLE_SMP
    pthread_mutex_unlock( &ssl_session_cache_lock);
#endif
    return -1;

}

void check_ssl_alert( request* req, int ret)
{
   int last_alert;

   if (ret == GNUTLS_E_WARNING_ALERT_RECEIVED || ret == GNUTLS_E_FATAL_ALERT_RECEIVED) 
   {
      last_alert = gnutls_alert_get(req->ssl_state);
      log_error_doc(req);
      fprintf(stderr, "Received alert \"%s\".\n", gnutls_alert_get_name(ret));
   }
}

int finish_handshake(request * current)
{
    int retval;

    retval = gnutls_handshake(current->ssl_state);

    if (retval == GNUTLS_E_AGAIN)
	retval = -1;
    else if (retval == GNUTLS_E_INTERRUPTED)
	retval = 1;
    else if (retval < 0) {
	if (gnutls_error_is_fatal(retval) != 0) {
	    log_error_doc(current);
	    fprintf(stderr, "TLS handshake error \"%s\"\n", gnutls_strerror(retval));
	    check_ssl_alert( current, retval);

	    /* we ignore the level of the alert, since we always
	     * send fatal alerts.
	     */
	    current->alert_to_send = gnutls_error_to_alert( retval, NULL);
	    if (current->alert_to_send >= 0) {
	    	current->status = SEND_ALERT;
	    }
	    retval = 1;
	} else {
	    check_ssl_alert( current, retval);
	    retval = 1;
	}
    } else if (retval == 0) {
	retval = 1;
	current->status = READ_HEADER;
    }

    return retval;
}

int send_alert(request * current)
{
    int retval;

    retval = gnutls_alert_send( current->ssl_state, 
    	GNUTLS_AL_FATAL, current->alert_to_send);

    if (retval == GNUTLS_E_AGAIN)
	retval = -1;
    else if (retval == GNUTLS_E_INTERRUPTED)
	retval = 1;
    else if (retval <= 0) {
        retval = 0;
	current->status = DEAD;
    }
    
    return retval;
}

#else /* a stub for initialize_ssl */

int initialize_ssl(void)
{
    log_error_time();
    fprintf(stderr, "SSL is not available in this build\n");
    exit(1);
}

#endif
